#include <CHM/CHMpreCompHead.h>
#pragma hdrstop

#include <CHM/CHMhl7ServerSocket.h>
#include <CHM/NETdll.h>

struct CHMhl7ServerSocketPrivate
{
   CHMhl7ServerSocketPrivate():Handle(0){}
   NETserverHandle Handle;
};

CHMhl7ServerSocket::CHMhl7ServerSocket()
{
   ensureInitialized();
   pMember = new CHMhl7ServerSocketPrivate();
   CHM_CHECK_CALL(NETserverCreate( this, &pMember->Handle ));
}

CHMhl7ServerSocket::~CHMhl7ServerSocket()
{
   NETuserHandle UserHandle;
   CHM_CHECK_CALL(NETserverDestroy( pMember->Handle, &UserHandle ));
   delete pMember;
}

/** Makes server listen on a specified port */
void CHMhl7ServerSocket::listen( short Port )
{
   CHM_CHECK_CALL(NETserverListen( pMember->Handle, Port ));
}

/** Stops the server from listening */
void CHMhl7ServerSocket::stopListening()
{
   CHM_CHECK_CALL(NETserverStopListening( pMember->Handle ));
}

/** Send a message. */
void CHMhl7ServerSocket::sendMessage( int TransportId, const char* Message )
{
   CHM_CHECK_CALL(NETtransportSendMessage( (NETconnectionHandle)TransportId, Message ));
}

/** Get the IP address */
CHMuint32 CHMhl7ServerSocket::getIpAddress( int ClientId )
{
   CHMuint32 Ip;
   CHM_CHECK_CALL(NETtransportGetIpAddress( (NETconnectionHandle)ClientId, &Ip ));
   return Ip;
}

/** Get the IP address as a string*/
const char* CHMhl7ServerSocket::getIpAddressString( int ClientId )
{
   const char* pIpAddress;
   CHM_CHECK_CALL(NETtransportGetIpAddressString( (NETconnectionHandle)ClientId, &pIpAddress ));
   return pIpAddress;
}

/** Gets the host name.*/
const char* CHMhl7ServerSocket::getHostName( int ClientId )
{
   const char* pHostName;
   CHM_CHECK_CALL(NETtransportGetHostName( (NETconnectionHandle)ClientId, &pHostName ));
   return pHostName;
}

/** Close a connection. */
void CHMhl7ServerSocket::closeConnection( int TransportId )
{
   CHM_CHECK_CALL(NETtransportCloseConnection( (NETconnectionHandle)TransportId ));
}

/** Close all connections. */
void CHMhl7ServerSocket::closeAllConnections()
{
   CHM_CHECK_CALL(NETserverCloseAllConnections( pMember->Handle ));
}

void CHMnetStop()
{
   CHM_CHECK_CALL(NETstop());
}

CHMhl7ClientSocket& CHMhl7ServerSocket::onNewConnection( int ClientId )
{ 
   return *( new CHMhl7ClientSocket( ClientId, true ) ); 
}

static void CHM_STD_CALL NETtransportOnConnect
(
   NETuserHandle UserHandle, 
   NETconnectionHandle TransportId 
)
{
   CHMhl7ClientSocket* pConnection = (CHMhl7ClientSocket*)UserHandle;
   pConnection->onConnect();
}

static void CHM_STD_CALL NETtransportOnMessage
( 
   NETuserHandle UserHandle, 
   NETuserHandle ServerUserHandle, 
   NETconnectionHandle TransportId, 
   const char* pMessage 
)
{
   CHMhl7ClientSocket* pConnection = (CHMhl7ClientSocket*)UserHandle;
   pConnection->onMessage( *pConnection, pMessage );
   if( ServerUserHandle )
   {
      CHMconnectionEvents* pServer = (CHMconnectionEvents*)ServerUserHandle;
      pServer->onMessage( *pConnection, pMessage );
   }
}

static void CHM_STD_CALL NETtransportOnClose
( 
   NETuserHandle UserHandle, 
   NETuserHandle ServerUserHandle, 
   NETconnectionHandle TransportId 
)
{
   //The onclose event is fired to the server before to the connection
   //this allows calls to the connection object to be valid before 
   //the connection is truly disconnected.
   CHMhl7ClientSocket* pConnection = (CHMhl7ClientSocket*)UserHandle;
   try
   {
      if( ServerUserHandle )
      {
         CHMconnectionEvents* pServer = (CHMconnectionEvents*)ServerUserHandle;
         pServer->onClose( *pConnection );
      }
   }
   catch(...) 
   {
      //its possible that if exceptions are thrown by the event handler, 
      //the connection is not closed, so this is our finally handler.
      pConnection->onClose( *pConnection );
      throw;
   }
   pConnection->onClose( *pConnection );
}

static void CHM_STD_CALL NETtransportOnError
( 
    NETuserHandle UserHandle, 
    NETuserHandle ServerUserHandle, 
    NETconnectionHandle TransportId, 
    int ErrorId, 
    const char* pErrorMessage 
)
{
   CHMhl7ClientSocket* pConnection = (CHMhl7ClientSocket*)UserHandle;
   //we handle this the same as the onCloseEvent
   try
   {
     if( ServerUserHandle )
     {
       CHMconnectionEvents* pServer = (CHMconnectionEvents*)ServerUserHandle;
       pServer->onTransportError( *pConnection, ErrorId, pErrorMessage );
     }
   }
   catch(...) 
   {
      //its possible that if exceptions are thrown by the event handler, 
      //the connection is not closed, so this is our finally handler.
      pConnection->onTransportError( *pConnection, ErrorId, pErrorMessage );
      throw;
   }
   pConnection->onTransportError( *pConnection, ErrorId, pErrorMessage );
}

static void CHM_STD_CALL NETserverOnError
( 
    NETuserHandle UserHandle, 
    NETserverHandle ServerHandle, 
    int ErrorId, 
    const char* pErrorMessage 
)
{
   CHMhl7ServerSocket* pServer = (CHMhl7ServerSocket*)UserHandle;
   pServer->onError( ErrorId, pErrorMessage );
}

static void CHM_STD_CALL NETserverOnNewConnection
( 
    NETuserHandle UserHandle, 
    NETserverHandle ServerHandle, 
    NETconnectionHandle TransportId, 
    NETuserHandle* pClientUserHandle
)
{
   CHMhl7ServerSocket* pServer = (CHMhl7ServerSocket*)UserHandle;
   *pClientUserHandle = (NETuserHandle)&pServer->onNewConnection( (int)TransportId );
}

void CHMhl7ServerSocket::ensureInitialized()
{
   static CHMboolean Initialized = false;
   if( !Initialized )
   {
      _NETsetTransportOnConnectCallback( NETtransportOnConnect );
      _NETsetTransportOnCloseCallback( NETtransportOnClose );
      _NETsetTransportOnMessageCallback( NETtransportOnMessage );
      _NETsetTransportOnErrorCallback( NETtransportOnError );
      _NETsetServerOnErrorCallback( NETserverOnError );
      _NETsetServerOnNewConnectionCallback( NETserverOnNewConnection );
      Initialized = true;
   }
}
